import math

from rest_framework import pagination
from rest_framework.response import Response
from rest_framework.utils.urls import replace_query_param


class Link_header_pagination( pagination.PageNumberPagination ):
    page_size_query_param = 'page_size'

    page_total_count_header = 'X-Pagination-Total-Count'
    page_count_header = 'X-Pagination-Page-Count'
    current_page_header = 'X-Pagination-Current-Page'
    size_page_header = 'X-Pagination-Per-Page'

    def get_paginated_response( self, data, request=None, serializer=None ):
        self.request = request or self.request
        self._data = data
        self._serializer = serializer

        self._generate_page_data()

        next_url = self.get_next_link()
        previous_url = self.get_previous_link()

        link = self.create_header_link( next_url, previous_url )

        headers = self.create_header( link )
        return self.create_response( data, headers )

    def create_header_link( self, next_url, previous_url ):
        if next_url is not None and previous_url is not None:
            link = '<{next_url}>; rel="next", <{previous_url}>; rel="prev"'
        elif next_url is not None:
            link = '<{next_url}>; rel="next"'
        elif previous_url is not None:
            link = '<{previous_url}>; rel="prev"'
        else:
            link = ''
        link = link.format( next_url=next_url, previous_url=previous_url )
        return link

    def create_header( self, link ):
        result = {
            self.page_total_count_header: self._data_count,
            self.page_count_header: min( self._page_size, len( self.data ) ),
            self.current_page_header: self._page_number + 1,
            self.size_page_header: self._page_size
        }
        if link:
            result[ 'Link' ] = link
        return result

    def create_response( self, data, headers ):
        if self._serializer is not None:
            serializer = self._serializer(
                data, many=True, context={ 'request': self.request } )
            return Response( serializer.data, headers=headers )
        else:
            return Response( data, headers=headers )

    def get_pages( self, data=None ):
        return math.ceil( self._data_count / self._page_size )

    def check_data( self, data ):
        return data

    def check_is_last_page( self ):
        return self._page_number in self.last_page_strings

    def _generate_page_data( self ):
        self.get_data_count()
        self._page_size = self.get_page_size( self.request )
        if self._page_size > 100:
            self._page_size = 100
        self._num_pages = self.get_pages( self._data )
        self._page_number = self.request.query_params.get(
            self.page_query_param, 0 )
        self._page_number = int( self._page_number )

        if self.check_is_last_page():
            self._page_number = self._num_pages

    def get_data_count( self ):
        try:
            self._data_count = self.page.paginator.count
        except:
            self._data_count = len( self.data )

    @property
    def data( self ):
        return self._data


class Paginate_search_es( Link_header_pagination ):
    def get_pages( self, data ):
        return math.ceil( data.count() / self._page_size )

    def create_response( self, data, headers ):
        if self._serializer is not None:
            serializer = self._serializer(
                self.data, many=True, context={ 'request': self.request } )
            return Response( serializer.data, headers=headers )
        else:
            data = self.data
            result_data = []
            for d in data:
                result_data.append( d.to_dict() )
            return Response( result_data, headers=headers )

    def get_data_count( self ):
        try:
            self._data_count = self.page.paginator.count
        except:
            self._data_count = self._data.count()

    def get_next_link( self ):
        if self._page_number >= self._num_pages:
            return None
        url = self.request.build_absolute_uri()
        page_number = self._page_number + 1
        return replace_query_param( url, self.page_query_param, page_number )

    def get_previous_link( self ):
        if self._page_number == 1:
            return None
        url = self.request.build_absolute_uri()
        page_number = self._page_number - 1
        return replace_query_param( url, self.page_query_param, page_number )

    @property
    def data( self ):
        try:
            return self.__data
        except AttributeError:
            start = self._page_size * ( self._page_number - 1  )
            if start < 0:
                start = 0
            end = start + self._page_size
            self.__data = self._data[ start:end ].execute()
            return self.__data



# auth0login/auth0backend.py

import requests
from social_core.backends.oauth import BaseOAuth2


class Auth0( BaseOAuth2 ):
    """Auth0 OAuth authentication backend"""
    name = 'auth0'
    SCOPE_SEPARATOR = ' '
    ACCESS_TOKEN_METHOD = 'POST'
    EXTRA_DATA = [
        ( 'picture', 'picture' )
    ]

    def authorization_url( self ):
        """Return the authorization endpoint."""
        return "https://" + self.setting( 'DOMAIN' ) + "/authorize"

    def access_token_url( self ):
        """Return the token endpoint."""
        return "https://" + self.setting( 'DOMAIN' ) + "/oauth/token"

    def get_user_id( self, details, response ):
        """Return current user id."""
        return details['user_id']

    def get_user_details( self, response ):
        url = 'https://' + self.setting( 'DOMAIN' ) + '/userinfo'
        headers = {'authorization': 'Bearer ' + response['access_token']}
        resp = requests.get( url, headers=headers )
        userinfo = resp.json()

        return {'username': userinfo['nickname'],
                'first_name': userinfo['name'],
                'picture': userinfo['picture'],
                'user_id': userinfo['sub']}



from django.test import TestCase

from chibi_user.models import User


class Test_user( TestCase ):
    def test_the_email_can_be_none( self ):
        user = User.objects.create_user(
            username='test', password='password', email=None )
        user_from_db = User.objects.get( pk=user.pk )
        self.assertEquals( user_from_db.pk, user.pk )

    def test_if_the_email_is_none_can_be_repeat( self ):
        user = User.objects.create_user(
            username='test', password='password', email=None )
        user_2 = User.objects.create_user(
            username='test_2', password='password', email=None )
        self.assertNotEqual( user.pk, user_2.pk )
        self.assertIsNone( user.email )
        self.assertIsNone( user_2.email )



from django.conf import settings
from django.db import models
from chibi import madness

AUTH_USER_MODEL = getattr( settings, 'AUTH_USER_MODEL', 'chibi_user.User' )


class Token( models.Model ):
    user = models.OneToOneField(
        AUTH_USER_MODEL, on_delete=models.CASCADE, related_name='token', )
    key = models.CharField( primary_key=True, max_length=64 )
    create_at = models.DateTimeField( auto_now_add=True )

    def save( self, *args, **kwargs ):
        if not self.key:
            self.key = self.generate_key()
        return super().save( *args, **kwargs )

    def __repr__(self):
        return "Token: %s" % self.key

    def __str__(self):
        return "Token %s" % self.key

    def generate_key( self, length=20 ):
        return madness.string.generate_token_b64( length )



import unittest
from chibi_django.exceptions import (
    Http_error, Http_internal_server_error, Http_bad_request, Http_not_found,
    generic_exception_handler )


class Test_generic_exception_handler( unittest.TestCase ):
    def test_should_solve_the_exception_http_error( self ):
        response = generic_exception_handler( Http_error(), None )
        self.assertTrue( response )
        self.assertEqual( response.data[ 'detail' ], 'Unhandled error.' )

    def test_should_solve_the_exception_http_internal_error( self ):
        response = generic_exception_handler(
            Http_internal_server_error(), None )
        self.assertTrue( response )
        self.assertEqual( response.data[ 'detail' ], 'Unhandled error.' )

    def test_should_solve_the_exception_http_bad_request( self ):
        response = generic_exception_handler( Http_bad_request(), None )
        self.assertTrue( response )
        self.assertEqual( response.data[ 'detail' ], 'Unhandled parameters.' )

    def test_should_solve_the_exception_http_not_found( self ):
        response = generic_exception_handler( Http_not_found(), None )
        self.assertTrue( response )
        self.assertEqual( response.data[ 'detail' ], 'Not found.' )



from django.contrib.auth.models import AbstractBaseUser, PermissionsMixin
# from django.utils.translation import ugettext_lazy as _
from django.db import models
from .token import Token
from chibi_user.managers import User_manager
from chibi_django.models import Chibi_model


class User( AbstractBaseUser, PermissionsMixin, Chibi_model ):
    """
    Modelo de usuarios para personalisar los campos
    """
    username = models.CharField( unique=True, max_length=64, )
    email = models.CharField(
        unique=True, null=True, default=None, max_length=64, )

    first_name = models.CharField( max_length=64, )
    last_name = models.CharField( max_length=64, )

    is_active = models.BooleanField( default=True, null=False )
    is_staff = models.BooleanField( default=False, null=False )
    is_superuser = models.BooleanField( default=False, null=False )

    objects = User_manager()
    USERNAME_FIELD = 'username'
    REQUIRED_FIELDS = []

    def __str__( self ):
        return "pk: {} :: username: {}".format( self.pk, self.username )

    def refresh_token( self ):
        """
        TODO: make test
        Refresca el token el usuario o lo crea

        Returns
        -------

        Token
            token que se genero para el usuario
        """
        try:
            if self.token:
                self.token.delete()
        except Token.DoesNotExist:
            pass
        finally:
            return Token.objects.create( user=self )



import jwt
import uuid
import warnings

from django.contrib.auth import get_user_model

from calendar import timegm
from datetime import datetime

from rest_framework_jwt.compat import get_username
from rest_framework_jwt.compat import get_username_field
from rest_framework_jwt.settings import api_settings


def get_secret_key( setting, payload=None ):
    """
    For enhanced security you may want to use a secret key based on user.

    This way you have an option to logout only this user if:
        - token is compromised
        - password is changed
        - etc.
    """
    if setting.JWT_GET_USER_SECRET_KEY:
        User = get_user_model()  # noqa: N806
        user = User.objects.get( pk=payload.get( 'user_id' ))
        key = str( setting.JWT_GET_USER_SECRET_KEY( user ))
        return key
    return setting.JWT_SECRET_KEY


def payload_handler( user ):
    username_field = get_username_field()
    username = get_username( user )

    warnings.warn(
        'The following fields will be removed in the future: '
        '`email` and `user_id`. ',
        DeprecationWarning
    )

    payload = {
        'user_id': user.pk,
        'username': username,
        'exp': datetime.utcnow() + api_settings.JWT_EXPIRATION_DELTA
    }
    if hasattr( user, 'email' ):
        payload['email'] = user.email
    if isinstance( user.pk, uuid.UUID ):
        payload['user_id'] = str( user.pk )

    payload[username_field] = username

    # Include original issued at time for a brand new token,
    # to allow token refresh
    if api_settings.JWT_ALLOW_REFRESH:
        payload['orig_iat'] = timegm(
            datetime.utcnow().utctimetuple()
        )

    if api_settings.JWT_AUDIENCE is not None:
        payload['aud'] = api_settings.JWT_AUDIENCE

    if api_settings.JWT_ISSUER is not None:
        payload['iss'] = api_settings.JWT_ISSUER

    return payload


def get_user_id_from_payload_handler( payload ):
    """
    Override this function if user_id is formatted differently in payload
    """
    warnings.warn(
        'The following will be removed in the future. '
        'Use `JWT_PAYLOAD_GET_USERNAME_HANDLER` instead.',
        DeprecationWarning
    )

    return payload.get( 'user_id' )


def get_username_from_payload_handler( payload ):
    """
    Override this function if username is formatted differently in payload
    """
    return payload.get( 'username' )


def encode_handler( payload, setting ):
    key = setting.JWT_PRIVATE_KEY or get_secret_key( setting, payload )
    return jwt.encode(
        payload, key, setting.JWT_ALGORITHM ).decode( 'utf-8' )


def decode_handler( token, setting ):
    options = {
        'verify_exp': setting.JWT_VERIFY_EXPIRATION,
    }
    # get user from token, BEFORE verification, to get user secret key
    unverified_payload = jwt.decode( token, None, False )
    secret_key = get_secret_key( setting, unverified_payload )

    return jwt.decode(
        token,
        setting.JWT_PUBLIC_KEY or secret_key,
        setting.JWT_VERIFY,
        options=options,
        leeway=setting.JWT_LEEWAY,
        audience=setting.JWT_AUDIENCE,
        issuer=setting.JWT_ISSUER,
        algorithms=[setting.JWT_ALGORITHM]
    )


def response_payload_handler( token, user=None, request=None ):
    """
    Returns the response data for both the login and refresh views.
    Override to return a custom response such as including the
    serialized representation of the User.

    Example:

    def response_payload_handler( token, user=None, request=None ):
        return {
            'token': token,
            'user': UserSerializer( user, context={'request': request}).data
        }

    """
    return {
        'token': token
    }



from rest_framework import serializers
from chibi_user.models import Token as Token_model, User as User_model
from chibi_django.serializers_fields import (
    parametrise_hyperlink_identity_field
)


class Token( serializers.ModelSerializer ):
    class Meta:
        model = Token_model
        fields = [ 'key', 'create_at' ]


class User( serializers.ModelSerializer ):
    token = Token( required=False )

    class Meta:
        model = User_model
        fields = [ 'pk', 'is_active', 'token' ]
        read_only_fields = [ 'pk', 'is_active', 'token' ]


class User_create( serializers.ModelSerializer ):
    url = parametrise_hyperlink_identity_field(
        lookup_obj_fields=( ( 'pk', 'pk', ), ),
        view_name='users:users-detail' )

    class Meta:
        model = User_model
        fields = [
            'pk', 'url', 'first_name', 'last_name', 'username', 'is_staff',
            'is_superuser', 'is_active' ]
        read_only_fields = [ 'pk', 'url' ]

    def create( self, validate_data ):
        user = User_model.objects.create( **validate_data )
        return user



from django.contrib.auth.models import BaseUserManager
from chibi import madness


class User_manager( BaseUserManager ):

    def create( self, *args, password='', **kw ):
        user = super().create( *args, password=password, **kw )
        user.set_password( password )
        user.save()
        user.refresh_token()
        return user

    def create_user( self, *args, **kw  ):
        """
        Crea un usuario sin privilegios especiales
        """
        user = self.create( *args, **kw )
        return user

    def create_superuser( self, *args, **kw ):
        """
        Crea un superusuario
        """
        user = self.create_user(
            *args, is_staff=True, is_superuser=True, **kw )
        return user

    def create_user_test( self, *args, username=None, password=None, **kw ):
        """
        Crea un usuario para pruebas
        """
        if not username:
            username = madness.string.generate_string()
        if not password:
            password = 'password'
        user = self.create_user(
            *args, username=username, password=password, **kw )
        return user

    def create_superuser_test(
            self, *args, username=None, password=None, **kw ):
        """
        Crea un super usuairo para las pruebas
        """
        if not username:
            username = madness.string.generate_string()
        if not password:
            password = 'password'
        user = self.create_superuser(
            *args, username=username, password=password, **kw )
        return user



# Generated by Django 2.1.5 on 2019-10-17 17:46

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('chibi_user', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='user',
            name='id',
            field=models.CharField(max_length=64, primary_key=True, serialize=False, verbose_name='Id'),
        ),
    ]



from django import http
from django.utils.translation import ugettext as _
from chibi_django.exceptions import Http_not_found, Http_internal_server_error
from rest_framework.renderers import JSONRenderer


def page_not_found(request, exception, template_name='404.html'):
    """
    404 error handler.
    Templates: :template:`404.html`
    Context: None
    """
    # format = request.META.get( 'CONTENT_TYPE', 'application/json' )
    base_exception = Http_not_found( _( "The resource does not exist." ) )
    result = base_exception.context
    result = JSONRenderer().render( result )

    return http.HttpResponseNotFound( result )


def server_error(request, template_name='500.html'):
    """
    500 error handler.
    Templates: :template:`500.html`
    Context: None
    """
    # format = request.META.get( 'CONTENT_TYPE', 'application/json' )
    base_exception = Http_internal_server_error()
    result = base_exception.context
    result = JSONRenderer().render( result )

    return http.HttpResponseServerError( result )



from django.test import TestCase
from chibi_user.factories import User as User_factory
from chibi_user.models import (
    Token as Token_model,
    User as User_model,
)


class Test_users( TestCase ):
    def test_factory( self ):
        user = User_factory.build()
        self.assertIsInstance( user, User_model )
        self.assertIsInstance( user.token, Token_model )



import logging
from rest_framework import status
from rest_framework.views import set_rollback
from rest_framework.response import Response
from rest_framework.views import exception_handler


logger = logging.getLogger( 'chibi_django.exceptions' )


class Http_error( Exception ):
    """
    exceptcion generica

    Arguments
    ---------
    status_code: int
    context: string, dict or list
    """
    def __init__(
            self, status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            context=None ):
        # if the context is a string return the format detail
        if context is None:
            context = 'Unhandled error.'
        if isinstance( context, ( str, ) ):
            context = { 'detail': context }
        self.status_code = status_code
        self.context = context


class Http_internal_server_error( Http_error ):
    """
    alias for the class py:class:`~Http_error`
    """
    pass


class Http_not_found( Http_error ):
    """
    excepcion for 404

    Arguments
    ---------
    context: string, dict or list
    """
    def __init__( self, context=None ):
        if context is None:
            context = 'Not found.'
        super().__init__(
            context=context, status_code=status.HTTP_404_NOT_FOUND )


class Http_bad_request( Http_error ):
    """
    Excepticion for 400

    Arguments
    ---------
    context: string, dict or list
    """
    def __init__( self, context=None ):
        if context is None:
            context = 'Unhandled parameters.'
        super().__init__(
            context=context, status_code=status.HTTP_400_BAD_REQUEST )


def generic_exception_handler( exc, context ):
    """
    maneja las excepciones genericas del api

    Parameters
    ----------
    exc: Exception
    contex: object

    Notes
    -----
    las exceptciones genericas pueden ser
        * :class:`~Http_error`
        * :class:`~Http_not_found`
        * :class:`~Http_internal_server_error`
        * :class:`~Http_bad_request`

    tambien soporta las excepciones por default de django
    """
    response = exception_handler( exc, context )

    if isinstance( exc, Http_error ):
        response = Response( exc.context, status=exc.status_code )
        set_rollback()

    if response is None and exc:
        response = Response(
            { "detail": "Unhandled error." },
            status=status.HTTP_500_INTERNAL_SERVER_ERROR )
        logger.exception( "unhandled error", exc_info=exc )
    return response



import factory
from factory import fuzzy
from faker import Factory as Faker_factory

from .token import Token
from chibi_user.models import User as User_model


faker = Faker_factory.create()


class User( factory.DjangoModelFactory ):
    username = factory.LazyAttribute( lambda x: faker.user_name() )
    is_active = fuzzy.FuzzyChoice( [ True, False ] )
    is_staff = fuzzy.FuzzyChoice( [ True, False ] )

    token = factory.RelatedFactory( Token, 'user' )

    class Meta:
        model = User_model



from datetime import datetime
from unittest.mock import Mock

from django.test import TestCase, override_settings
from rest_framework import status

from chibi_user.middleware.metric import User_for_api


@override_settings( ROOT_URLCONF='chibi_user.tests' )
class metric( TestCase ):

    def test_save_start_time( self ):
        request = Mock()
        get_response = Mock()
        middleware = User_for_api( get_response=get_response )
        middleware( request )
        self.assertTrue( hasattr( request, '_start_time' ) )
        self.assertIsInstance( request._start_time, datetime )
        return middleware, request

    def test_check_for_request_anonymous( self ):
        middleware, request = self.test_save_start_time()
        del request.user
        metrics = {}
        middleware.set_user_to_metrics( request, metrics )
        self.assertTrue( metrics[ 'is_anonymous' ] )

    def test_check_for_request_with_user( self ):
        middleware, request = self.test_save_start_time()
        request.user = Mock()
        metrics = {}
        middleware.set_user_to_metrics( request, metrics )
        self.assertFalse( metrics[ 'is_anonymous' ] )
        self.assertIn( 'user_pk', metrics )
        self.assertEqual( metrics[ 'user_pk' ], request.user.pk )

    def test_no_error_for_the_user( self ):
        middleware, request = self.test_save_start_time()
        request.user = Mock()
        response = Mock()
        response.status_code = status.HTTP_404_NOT_FOUND
        metrics = {}
        middleware.get_url_name_from_response = Mock()
        middleware.get_url_name_from_response.return_value = 'some_name'
        middleware.is_an_error_from_user( request, response, metrics )
        self.assertFalse( metrics[ 'user_error' ] )

    def test_error_for_the_user( self ):
        middleware, request = self.test_save_start_time()
        request.user = Mock()
        response = Mock()
        response.status_code = status.HTTP_404_NOT_FOUND
        metrics = {}
        middleware.get_url_name_from_response = Mock()
        middleware.get_url_name_from_response.return_value = 'some_name-detail'
        middleware.is_an_error_from_user( request, response, metrics )
        self.assertTrue( metrics[ 'user_error' ] )



from __future__ import unicode_literals

from django.utils.translation import ugettext_lazy as _
from rest_framework import exceptions
from rest_framework.authentication import (
    get_authorization_header, BaseAuthentication
)

from chibi_user.models import Token


__all__ = [ 'Token_simple_authentication' ]


class Token_simple_authentication( BaseAuthentication ):
    model = Token

    def authenticate( self, request ):
        auth = get_authorization_header( request ).split( b';' )
        if not auth:
            return None
        if len( auth ) == 1 and not auth[0]:
            return None

        auth = self.parse_tokens( auth )

        if len( auth ) == 0:
            msg = _( 'Invalid token header. No credentials provided.' )
            raise exceptions.AuthenticationFailed( msg )

        return self.authenticate_credentials( auth, request )

    def authenticate_credentials( self, tokens, request ):
        try:
            key = tokens.pop( 'Token' )
            token = self.model.objects.select_related( 'user' ).get( key=key )
        except ( KeyError, self.model.DoesNotExist ):
            raise exceptions.AuthenticationFailed( _( 'Invalid token.' ) )

        if not token.user.is_active:
            raise exceptions.AuthenticationFailed(
                _( 'User inactive or deleted.' ) )

        return ( token.user, token )

    def authenticate_header( self, request ):
        return 'Token'

    def parse_tokens( self, tokens ):
        result = {}
        for token in tokens:
            if len( token ) == 0:
                msg = _( 'Invalid token header. No credentials provided.' )
                raise exceptions.AuthenticationFailed( msg )
            s_token = token.split()
            len_s_token = len( s_token )
            if len_s_token == 1 or len_s_token > 2:
                msg = _(
                    'Invalid token header. Token string '
                    'should not contain spaces.' )
                raise exceptions.AuthenticationFailed( msg )
            try:
                result[ s_token[0].decode() ] = s_token[1].decode()
            except UnicodeError:
                msg = _(
                    'Invalid token header. Token string '
                    'should not contain invalid characters.' )
                raise exceptions.AuthenticationFailed( msg )
        return result

    def check_prefix( self, prefix ):
        return 'token' if prefix == 'token' else None


'''
import jwt
from rest_framework_jwt.settings import api_settings
from django.contrib.auth import get_user_model
from django.utils.encoding import smart_text
from rest_framework_jwt.authentication import BaseJSONWebTokenAuthentication

jwt_decode_handler = api_settings.JWT_DECODE_HANDLER
jwt_get_username_from_payload = api_settings.JWT_PAYLOAD_GET_USERNAME_HANDLER


class Base_JWT_authentication( BaseAuthentication ):
    """
    Token based authentication using the JSON Web Token standard.
    """

    def authenticate( self, request ):
        """
        Returns a two-tuple of `User` and token if a valid signature has been
        supplied using JWT-based authentication.  Otherwise returns `None`.
        """
        jwt_value, setting = self.get_jwt_value( request )
        if jwt_value is None:
            return None

        try:
            payload = setting.jwt_decode_handler( jwt_value )
        except jwt.ExpiredSignature:
            msg = _( 'Signature has expired.' )
            raise exceptions.AuthenticationFailed( msg )
        except jwt.DecodeError:
            msg = _( 'Error decoding signature.' )
            raise exceptions.AuthenticationFailed( msg )
        except jwt.InvalidTokenError:
            raise exceptions.AuthenticationFailed()

        user = self.authenticate_credentials( payload )

        return ( user, payload )

    def authenticate_credentials( self, payload, setting ):
        """
        Returns an active user that matches the payload's user id and email.
        """
        User = get_user_model()
        username = setting.jwt_get_username_from_payload( payload )

        if not username:
            msg = _( 'Invalid payload.' )
            raise exceptions.AuthenticationFailed( msg )

        try:
            user = User.objects.get_by_natural_key( username )
        except User.DoesNotExist:
            msg = _( 'Invalid signature.' )
            raise exceptions.AuthenticationFailed( msg )

        if not user.is_active:
            msg = _( 'User account is disabled.' )
            raise exceptions.AuthenticationFailed( msg )

        return user

    def get_jwt_value( self, request ):
        auth = get_authorization_header( request ).split()
        auth_header_prefix = api_settings.JWT_AUTH_HEADER_PREFIX.lower()

        if not auth:
            if api_settings.JWT_AUTH_COOKIE:
                return request.COOKIES.get( api_settings.JWT_AUTH_COOKIE )
            return None, None

        auth_settings = self.find_settings( auth )
        if not auth_settings:
            return None, None

        if len( auth ) == 1:
            msg = _( 'Invalid Authorization header. No credentials provided.' )
            raise exceptions.AuthenticationFailed( msg )
        elif len( auth ) > 2:
            msg = _(
                'Invalid Authorization header. Credentials string '
                'should not contain spaces.' )
            raise exceptions.AuthenticationFailed( msg )

        return auth[1], auth_settings


class JSONWebTokenAuthentication( BaseJSONWebTokenAuthentication ):
    """
    Clients should authenticate by passing the token key in the "Authorization"
    HTTP header, prepended with the string specified in the setting
    `JWT_AUTH_HEADER_PREFIX`. For example:

        Authorization: JWT eyJhbGciOiAiSFMyNTYiLCAidHlwIj
    """
    www_authenticate_realm = 'api'

    def get_jwt_value( self, request ):
        auth = get_authorization_header( request ).split()
        auth_header_prefix = api_settings.JWT_AUTH_HEADER_PREFIX.lower()

        if not auth:
            if api_settings.JWT_AUTH_COOKIE:
                return request.COOKIES.get( api_settings.JWT_AUTH_COOKIE )
            return None

        auth_settings = self.find_settings( auth )
        if not auth_settings:
            return None

        if len( auth ) == 1:
            msg = _( 'Invalid Authorization header. No credentials provided.' )
            raise exceptions.AuthenticationFailed( msg )
        elif len( auth ) > 2:
            msg = _(
                'Invalid Authorization header. Credentials string '
                'should not contain spaces.' )
            raise exceptions.AuthenticationFailed( msg )

        return auth[1], auth_settings

    def find_settings( self, auth ):
        auth_realm = smart_text( auth[0].lower() )

        for realm, setting in api_settings.jwt_multi_realm.items():
            if auth_realm == realm.lower():
                return setting
        return None

    def authenticate_header( self, request ):
        """
        Return a string to be used as the value of the `WWW-Authenticate`
        header in a `401 Unauthenticated` response, or `None` if the
        authentication scheme should return `403 Permission Denied` responses.
        """
        return '{0} realm="{1}"'.format(
            api_settings.JWT_AUTH_HEADER_PREFIX, self.www_authenticate_realm )
'''



import datetime
from django.conf import settings
from rest_framework.settings import APISettings


USER_SETTINGS = getattr(settings, 'JWT_AUTH', None)

DEFAULTS = {
    'JWT_ENCODE_HANDLER': 'chibi_user.jwt.encode_handler',
    'JWT_DECODE_HANDLER': 'chibi_user.jwt.decode_handler',
    'JWT_PAYLOAD_HANDLER': 'chibi_user.jwt.payload_handler',

    'JWT_PAYLOAD_GET_USER_ID_HANDLER':
        'chibi_user.jwt.get_user_id_from_payload_handler',

    'JWT_PRIVATE_KEY': None,
    'JWT_PUBLIC_KEY': None,

    'JWT_PAYLOAD_GET_USERNAME_HANDLER':
        'chibi_user.jwt.get_username_from_payload_handler',

    'JWT_RESPONSE_PAYLOAD_HANDLER':
        'chibi_user.jwt.response_payload_handler',

    'JWT_SECRET_KEY': settings.SECRET_KEY,
    'JWT_GET_USER_SECRET_KEY': None,
    'JWT_ALGORITHM': 'HS256',
    'JWT_VERIFY': True,
    'JWT_VERIFY_EXPIRATION': True,
    'JWT_LEEWAY': 0,
    'JWT_EXPIRATION_DELTA': datetime.timedelta( seconds=300 ),
    'JWT_AUDIENCE': None,
    'JWT_ISSUER': None,

    'JWT_ALLOW_REFRESH': False,
    'JWT_REFRESH_EXPIRATION_DELTA': datetime.timedelta( days=7 ),

    'JWT_AUTH_HEADER_PREFIX': 'JWT',
    'JWT_AUTH_COOKIE': None,
}

# List of settings that may be in string import notation.
IMPORT_STRINGS = (
    'JWT_ENCODE_HANDLER',
    'JWT_DECODE_HANDLER',
    'JWT_PAYLOAD_HANDLER',
    'JWT_PAYLOAD_GET_USER_ID_HANDLER',
    'JWT_PAYLOAD_GET_USERNAME_HANDLER',
    'JWT_RESPONSE_PAYLOAD_HANDLER',
    'JWT_GET_USER_SECRET_KEY',
)

api_settings = APISettings(USER_SETTINGS, DEFAULTS, IMPORT_STRINGS)



import logging
from django.utils.translation import gettext as _
from chibi_django import mixins
from rest_framework import viewsets

from django.core.exceptions import ValidationError
from django.db.models.query import QuerySet
from django.http import Http404
from django.shortcuts import get_object_or_404 as _get_object_or_404

from rest_framework.settings import api_settings
from elasticsearch.exceptions import NotFoundError


logger = logging.getLogger( 'chibi_django.exceptions' )


def get_object_or_404( queryset, *filter_args, **filter_kwargs ):
    try:
        index = queryset.index()
        if len( index._doc_type ) > 1:
            logger.warning( f'se encontro mas de un modelo {index._doc_type}' )
        return index._doc_type[0].get( filter_kwargs[ 'pk' ] )
    except ( TypeError, ValueError, ValidationError, NotFoundError ):
        raise Http404

class Multi_permission_viewset( viewsets.GenericViewSet ):
    def get_permissions(self):
        """
        Instantiates and returns the list of permissions that this view requires.
        """
        func = getattr( self, self.action )
        if hasattr( func, 'permission_classes' ):
            return [ permission() for permission in func.permission_classes ]
        return [permission() for permission in self.permission_classes]


class Multi_serializer_viewset( Multi_permission_viewset ):
    def get_serializer( self, *args, serializer_name='default', **kw ):
        serializer_class = self.get_serializer_class( serializer_name )
        kw[ 'context' ] = self.get_serializer_context()
        return serializer_class( *args, **kw )

    def get_serializer_class( self, serializer_name='default' ):
        serializers = super().get_serializer_class()
        if not isinstance( serializers, dict ):
            return serializers
        try:
            serializer_class = serializers[ serializer_name ]
        except KeyError:
            try:
                serializer_class = serializers[ 'default' ]
            except KeyError:
                raise KeyError( _(
                    'The serializer %s is not defined' ) % serializer_name )
        return serializer_class


class Nested_view_set( Multi_serializer_viewset ):
    def create_build_data( self, data ):
        return data

    def update_build_data( self, data ):
        return data


class Elastic_view_set( Nested_view_set ):
    def get_object( self ):
        queryset = self.filter_queryset( self.get_queryset() )

        # Perform the lookup filtering.
        lookup_url_kwarg = self.lookup_url_kwarg or self.lookup_field

        assert lookup_url_kwarg in self.kwargs, (
            'Expected view %s to be called with a URL keyword argument '
            'named "%s". Fix your URL conf, or set the `.lookup_field` '
            'attribute on the view correctly.' %
            ( self.__class__.__name__, lookup_url_kwarg )
        )

        filter_kwargs = {self.lookup_field: self.kwargs[lookup_url_kwarg]}
        obj = get_object_or_404( queryset, **filter_kwargs )

        # May raise a permission denied
        # self.check_object_permissions( self.request, obj )

        return obj


class Model_viewset(
        mixins.Create_model, mixins.Retrieve_model, mixins.Update_model,
        mixins.Destroy_model, mixins.List_model, Nested_view_set ):
    pass


class Elastic_model_viewset(
        mixins.Elastic_list_model, mixins.Elastic_retrieve_model,
    mixins.Elastic_create_model, Elastic_view_set ):
    pass


class Read_only_model_viewset(
        mixins.Retrieve_model, mixins.List_model, Nested_view_set ):
    pass



from factory import fuzzy
from faker import Factory as Faker_factory
from chibi_user.models import Token as Token_model
from datetime import date
import factory

from chibi import madness


faker = Faker_factory.create()
start_date = date( 2016, 1, 1 )


class Token( factory.DjangoModelFactory ):
    user = factory.SubFactory( 'chibi_user.factories.User' )
    key = factory.lazy_attribute(
        lambda x: madness.string.generate_token( 20 ) )
    create_at = fuzzy.FuzzyDate( start_date )

    class Meta:
        model = Token_model



from rest_framework import status
from rest_framework.response import Response
from rest_framework.settings import api_settings


class Location_header:
    def get_success_headers( self, data ):
        try:
            return { 'Location': data[ api_settings.URL_FIELD_NAME ] }
        except ( TypeError, KeyError ):
            return {}


class Create_model( Location_header ):
    def create( self, request, *args, **kw ):
        serializer = self.get_serializer(
            data=self.create_build_data( request.data ),
            serializer_name='create' )
        serializer.is_valid( raise_exception=True )
        self.perform_create( serializer )
        headers = self.get_success_headers( serializer.data )
        return Response( status=status.HTTP_201_CREATED, headers=headers )

    def perform_create( self, serializer ):
        serializer.save()


class List_model:
    def list( self, request, *args, **kw ):
        queryset = self.filter_queryset( self.get_queryset() )

        page = self.paginate_queryset( queryset )
        if page is not None:
            serializer = self.get_serializer(
                page, many=True, serializer_name='list' )
            return self.get_paginated_response( serializer.data )

        serializer = self.get_serializer(
            queryset, many=True, serializer_name='list' )
        return Response( serializer.data )


class Retrieve_model:
    def retrieve( self, request, *args, **kw ):
        instance = self.get_object()
        serializer = self.get_serializer(
            instance, serializer_name='retrieve' )
        return Response( serializer.data )


class Update_model:
    def update( self, request, *args, **kw ):
        partial = kw.pop( 'partial', False )
        instance = self.get_object()
        serializer = self.get_serializer(
            instance, data=self.update_build_data( request.data ),
            partial=partial, serializer_name='update' )

        serializer.is_valid( raise_exception=True )
        self.perform_update( serializer )
        return Response( status=status.HTTP_204_NO_CONTENT )
        # return Response( serializer.data )

    def perform_update( self, serializer ):
        serializer.save()

    def partial_update( self, request, *args, **kwargs ):
        kwargs['partial'] = True
        return self.update( request, *args, **kwargs )


class Destroy_model:
    def destroy( self, request, *args, **kw ):
        instance = self.get_object()
        self.perform_destroy( instance )
        return Response( status=status.HTTP_204_NO_CONTENT )

    def perform_destroy( self, instance ):
        instance.delete()


class Elastic_list_model:
    def list( self, request, *args, **kw ):
        queryset = self.filter_queryset( self.get_queryset() )
        serializer = self.get_serializer_class( 'list' )

        page = self.pagination_class()
        return page.get_paginated_response(
            queryset, request, serializer=serializer )

class Elastic_retrieve_model:
    def retrieve( self, request, *args, **kw ):
        instance = self.get_object()
        serializer = self.get_serializer(
            instance, serializer_name='retrieve' )
        return Response( serializer.data )


class Elastic_create_model( Location_header ):
    def create( self, request, *args, **kw ):
        serializer = self.get_serializer(
            data=self.create_build_data( request.data ),
            serializer_name='create' )
        serializer.is_valid( raise_exception=True )
        self.perform_create( serializer )
        headers = self.get_success_headers( serializer.data )
        return Response( status=status.HTTP_201_CREATED, headers=headers )

    def perform_create( self, serializer ):
        serializer.save()



from django.test import TestCase
from unittest.mock import patch
from chibi_user.models import User


class Test_user_manager( TestCase ):
    @patch( "chibi_user.models.User.save" )
    @patch( "chibi_user.models.Token.save" )
    def test_create_user( self, token_save, user_save ):
        user = User.objects.create_user( username='test', password='password' )
        self.assertTrue( user_save.called )
        self.assertTrue( token_save.called )
        self.assertEqual( user.username, "test" )
        self.assertEqual( user._password, "password" )
        self.assertTrue( user.is_active )
        self.assertFalse( user.is_staff )

    @patch( "chibi_user.models.User.save" )
    @patch( "chibi_user.models.Token.save" )
    def test_create_superuser( self, user_save, token_save):
        user = User.objects.create_superuser(
            username='test', password='password' )
        self.assertTrue( user.is_staff )
        self.assertTrue( user.is_superuser )
        self.assertTrue( user_save.called )
        self.assertTrue( token_save.called )

    @patch( "chibi_user.models.User.save" )
    @patch( "chibi_user.models.Token.save" )
    def test_create_superuser_test( self, user_save, token_save ):
        user = User.objects.create_superuser_test()
        self.assertTrue( user.is_staff )
        self.assertTrue( user.is_superuser )
        self.assertTrue( user_save.called )
        self.assertTrue( token_save.called )

    @patch( "chibi_user.models.User.save" )
    @patch( "chibi_user.models.Token.save" )
    def test_create_user_test( self, user_save, token_save ):
        user = User.objects.create_user_test()
        self.assertFalse( user.is_staff )
        self.assertFalse( user.is_superuser )
        self.assertTrue( user_save.called )
        self.assertTrue( token_save.called )



from django.db import models
from django.test import TestCase, override_settings
from rest_framework import status, serializers
from rest_framework.permissions import AllowAny
from rest_framework.reverse import reverse
from rest_framework.test import APIClient, APIRequestFactory
from rest_framework_nested import routers

from chibi_django.view_set import Model_viewset


factory = APIRequestFactory()


class Action( models.Model ):
    pass


class Action_serializer( serializers.ModelSerializer ):
    class Meta:
        model = Action
        fields = ( 'pk', )


class Mock_View( Model_viewset ):
    authentication_classes = []
    permission_classes = ( AllowAny, )
    queryset = Action.objects.all()
    serializer_class = Action_serializer


router = routers.SimpleRouter()
router.register( r'', Mock_View, basename='mock' )

urlpatterns = router.urls


@override_settings(
    ROOT_URLCONF='chibi_django.tests.integration.view_set' )
class Test_view_set( TestCase ):

    def test_list_should_return_200( self ):
        self.client = APIClient( enforce_csrf_checks=True )
        url = reverse( 'mock-list' )
        response = self.client.get( url )
        self.assertEqual( response.status_code, status.HTTP_200_OK )

    def test_list_create_should_return_201( self ):
        self.client = APIClient( enforce_csrf_checks=True )
        url = reverse( 'mock-list' )
        response = self.client.post( url )
        self.assertEqual( response.status_code, status.HTTP_201_CREATED )

    def test_detail_should_return_200( self ):
        self.client = APIClient( enforce_csrf_checks=True )
        a = Action.objects.create()
        url = reverse( 'mock-detail', kwargs={ 'pk': a.pk } )
        response = self.client.get( url )
        self.assertEqual( response.status_code, status.HTTP_200_OK )

    def test_delete_should_return_204( self ):
        self.client = APIClient( enforce_csrf_checks=True )
        a = Action.objects.create()
        url = reverse( 'mock-detail', kwargs={ 'pk': a.pk } )
        response = self.client.delete( url )
        self.assertEqual( response.status_code, status.HTTP_204_NO_CONTENT )

    def test_path_should_return_204( self ):
        self.client = APIClient( enforce_csrf_checks=True )
        a = Action.objects.create()
        url = reverse( 'mock-detail', kwargs={ 'pk': a.pk } )
        response = self.client.patch( url )
        self.assertEqual( response.status_code, status.HTTP_204_NO_CONTENT )

    def test_put_should_return_204( self ):
        self.client = APIClient( enforce_csrf_checks=True )
        a = Action.objects.create()
        url = reverse( 'mock-detail', kwargs={ 'pk': a.pk } )
        response = self.client.put( url )
        self.assertEqual( response.status_code, status.HTTP_204_NO_CONTENT )



# vim: set fileencoding=utf-8 :
from django.contrib import admin
from django.contrib.auth.admin import UserAdmin
from django.contrib.auth.forms import UserChangeForm

from chibi_user.models import User


class User_admin_change_form( UserChangeForm ):
    class Meta( UserChangeForm.Meta ):
        model = User


class User_admin( UserAdmin ):
    form = User_admin_change_form


admin.site.register( User, User_admin )



from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APIClient

from chibi_user.models import Token
from chibi_user.tests import get_user_test, get_superuser_test
from test_runners.snippet.response import get_location, assert_status_code


fake = Faker_factory.create()


class Test_views_normal_user( TestCase ):
    model = Token
    path = '/token/'

    def setUp( self ):
        self.client = APIClient( enforce_csrf_checks=True )
        self.super_user, self.super_token  = get_superuser_test()
        self.user, self.user_token = get_user_test()
        self.client.credentials( HTTP_AUTHORIZATION=str( self.user_token ) )

    def test_fail_with_normal_user( self ):
        response = self.client.get(
            '/users/', HTTP_AUTHORIZATION=str( self.user_token ) )

        assert_status_code( response, status.HTTP_403_FORBIDDEN )


class Test_views( TestCase ):
    model = Token
    path = '/token/'

    def setUp( self ):
        self.client = APIClient( enforce_csrf_checks=True )
        self.super_user, self.super_token  = get_superuser_test()
        self.user, self.user_token = get_superuser_test()
        self.client.credentials( HTTP_AUTHORIZATION=str( self.user_token ) )

    def test_access_with_super_user( self ):
        auth = str( self.super_token )
        response = self.client.get( '/users/',
                                    HTTP_AUTHORIZATION=auth )

        self.assertEqual( response.status_code, status.HTTP_200_OK )

    def test_create_user( self ):
        auth = str( self.super_token )
        response = self.client.post(
            '/users/', HTTP_AUTHORIZATION=auth,
            data={
                'first_name': fake.first_name(),
                'last_name': fake.last_name(), 'email': fake.email(),
                'username': fake.user_name(),
            } )

        self.assertEqual( response.status_code, status.HTTP_201_CREATED,
                          ( "the status code should be 200 instead "
                            "of {}\ndata:{}" ).format( response.status_code,
                                                       response.data ) )
        assert_status_code( response, status.HTTP_201_CREATED )
        response = get_location( response, client=self.client )

        self.assertIsInstance( response.data[ 'pk' ], str )
        self.assertIn( 'token', response.data )
        self.assertIn( 'key', response.data[ 'token' ] )
        self.assertIsInstance( response.data[ 'token' ][ 'key' ], str )
        return response.data

    def test_delete_user( self ):
        data = self.test_create_user()
        auth = str( self.super_token )
        url = reverse( 'users:users-detail', kwargs={ 'pk': data[ 'pk' ] } )
        response = self.client.delete( url, HTTP_AUTHORIZATION=auth )

        self.assertEqual( response.status_code, status.HTTP_204_NO_CONTENT )

    def test_refresh_token( self ):
        data = self.test_create_user()
        auth = str( self.super_token )
        url = reverse(
            'users:users-refresh-token', kwargs={ 'pk': data[ 'pk' ] } )
        response = self.client.post( url, HTTP_AUTHORIZATION=auth )

        self.assertEqual( response.status_code, status.HTTP_200_OK )
        self.assertNotEqual( data[ 'token' ][ 'key' ],
                             response.data[ 'key' ] )



from django.db import models


class Base64_field( models.CharField ):

    def pre_save( self, model_instance, add ):
        string = super().pre_save( model_instance, add )
        return string

    def from_db_value(self, value, expression, connection, context):
        if value is None:
            return value
        return value

    def to_python(self, value):
        if value is None:
            return value

        return value



from rest_framework import serializers
from rest_framework.reverse import reverse
from marshmallow import fields


class Impossible_build_url( Exception ):
    pass


class parametrise_hyperlink_identity_field(
        serializers.HyperlinkedIdentityField ):

    lookup_fields = None
    lookup_obj_fields = None
    lookup_request_path = None

    def __init__( self, *args, **kargs ):
        self.lookup_fields = kargs.pop( 'lookup_fields', self.lookup_fields )
        self.lookup_obj_fields = kargs.pop(
            'lookup_obj_fields', self.lookup_obj_fields)
        self.lookup_request_path = kargs.pop(
            'lookup_request_path', self.lookup_request_path )
        super().__init__( *args, **kargs )

    def get_url( self, obj, view_name, request, format ):
        kargs = {}

        try:
            self._build_kargs_from_model( kargs )
            self._build_kargs_from_obj( obj, kargs )
            self._build_kargs_from_request( request, kargs )
        except Impossible_build_url:
            return None

        url = reverse( view_name, kwargs=kargs, request=request,
                       format=format )
        return url

    def _build_kargs_from_model( self, kargs ):
        if self.lookup_fields is not None:
            raise NotImplementedError(
                "the lookup_fields are for get the parameter of the model"
            )

    def _build_kargs_from_obj( self, obj, kargs ):
        if self.lookup_obj_fields is not None:
            if isinstance( obj, dict ):
                for lookup, field in self.lookup_obj_fields:
                    value = obj[ field ]
                    if value is None:
                        msg = (
                            "cannot find the field `{}` in the object"
                        ).format( field )
                        raise Impossible_build_url( msg )
                    kargs[ lookup ] = value
            else:
                for lookup, field in self.lookup_obj_fields:
                    if '.' in field:
                        value = obj
                        for part_field in field.split( '.' ):
                            value = getattr( value, part_field )
                    else:
                        value = getattr( obj, field )
                    if value is None:
                        msg = (
                            "cannot find the field `{}` in the object"
                        ).format( field )
                        raise Impossible_build_url( msg )
                    kargs[ lookup ] = value

    def _build_kargs_from_request( self, request, kargs, ):
        if self.lookup_request_path is not None:
            path_kargs = request.resolver_match.kwargs
            for lookup, field in self.lookup_request_path:
                value = path_kargs[ lookup ]
                if value is None:
                    msg = (
                        "cannot find the field '{}' in the request path"
                    ).format( field )
                    raise Impossible_build_url( msg )
                kargs[ lookup ] = value

    def to_internal_value( self, *args, **kw ):
        result = super().to_internal_value( *args, **kw )
        return result

    def to_representation( self, *args, **kw ):
        result = super().to_representation( *args, **kw )
        return str( result )

    def serialize( self, *args, **kw ):
        import pdb
        pdb.set_trace()
        return self.to_internal_value( self, *args, **kw )
        return super().serialize( *args, **kw  )

    def deserialize( self, *args, **kw ):
        return self.to_representation( self, *args, **kw )
        return super().deserialize( *args, **kw  )



from rest_framework import status
from test_runners import status_code as status_test


class Test_unauthorized( status_test.Test_status_code ):
    expected_status_code = status.HTTP_401_UNAUTHORIZED

    def setUp( self ):
        token = self.get_token()
        if token is not None:
            self.client.credentials( HTTP_AUTHORIZATION=str( token ) )

    def get_token( self ):
        pass


class Test_list( Test_unauthorized, status_test.Test_list ):
    pass


class Test_retrieve( Test_unauthorized, status_test.Test_retrieve ):
    pass


class Test_create( Test_unauthorized, status_test.Test_create ):
    pass



import json
import logging

from chibi import madness
from django.db import models
from django.db.models.signals import pre_save
from django.dispatch import receiver
from django.utils.translation import ugettext_lazy as _


logger = logging.getLogger( 'chibi_django.models' )


class Chibi_model( models.Model ):
    id = models.CharField( _( "Id" ), max_length=64, primary_key=True )

    class Meta:
        abstract = True


@receiver( pre_save )
def chibi_model_create_pk( sender, instance, **kw ):
    if not instance.pk and issubclass( sender, Chibi_model ):
        model_meta = sender._meta
        max_length = model_meta.get_field( 'id' ).max_length
        start = max_length // 2
        max_retry = 5
        count_retry = 0
        while True:
            try_pk = madness.string.generate_token_b64( length=start )
            is_pk_exists = sender.objects.filter( pk=try_pk ).count()
            if is_pk_exists == 0:
                break
            logger.warinig(
                "colicion de pks",
                extra={
                    'number_pk_collide': 1, 'pk_collide': [ try_pk ],
                    'length_of_pk': start, 'count_of_retry': count_retry,
                    'max_length_pk': max_length,
                    'current_length': len( try_pk )
                } )
            if count_retry >= max_retry:
                start += 1
                count_retry += 1
                start = min( start, max_length )

        instance.pk = try_pk


class Header( models.Model ):
    key = models.CharField( max_length=128 )
    value_str = models.TextField( blank=True )

    class Meta:
        abstract = True

    def save( self, *args, **kw ):
        self.value = self.value
        super().save( *args, **kw )

    @property
    def value( self ):
        try:
            value = self._value
        except AttributeError:
            try:
                value = json.loads( self.value_str )
            except ( ValueError, TypeError ):
                value = self.value_str
        return value

    @value.setter
    def value( self, value ):
        if isinstance( value, str ):
            self.value_str = value
            self._value = value
        else:
            self.value_str = json.dumps( value )
            self._value = value

    def __str__( self ):
        return "{}: {}".format( self.key, self.value )

    def __repr__( self ):
        return "Header( key='{}' value={} )".format( self.key, self.value )

    def __eq__( self, other ):
        if not isinstance( other, Header ):
            return False
        return self.key == other.key and self.value == self.value



from rest_framework import permissions, viewsets, status
from rest_framework.response import Response
from rest_framework.test import APIRequestFactory
from rest_framework_nested import routers

try:
    from chibi_user.models import User as User_model, Token as Token_model
except Exception:
    pass


factory = APIRequestFactory()


def get_superuser_test():
    try:
        user = User_model.objects.filter( is_staff=True )[0]
    except:
        user = User_model.objects.create_superuser_test()
    try:
        token = user.token
    except Token_model.DoesNotExist:
        token = user.refresh_token()
    return user, token


def get_user_test( pk=None ):
    try:
        user = User_model.objects.filter( is_staff=False )[0]
    except:
        user = User_model.objects.create_user_test( pk=pk )
    try:
        token = user.token
    except Token_model.DoesNotExist:
        token = user.refresh_token()
    return user, token


class Mock_View_authenticated( viewsets.ViewSet ):
    permission_classes = ( permissions.IsAuthenticated, )

    def list( self, request, format=None ):
        return Response(
            { 'a': 1, 'b': 2, 'c': 3 }, status=status.HTTP_200_OK )

    def retrieve( self, request, pk, format=None ):
        return Response(
            { 'a': 1, 'b': 2, 'c': 3 }, status=status.HTTP_200_OK )

    def update( self, request, pk, format=None ):
        return Response(
            { 'a': 1, 'b': 2, 'c': 3 }, status=status.HTTP_200_OK )

    def create( self, request, format=None ):
        return Response(
            { 'a': 1, 'b': 2, 'c': 3 }, status=status.HTTP_200_OK )


class Mock_view_all_404( viewsets.ViewSet ):
    permission_classes = ( permissions.IsAuthenticated, )

    def list( self, request, format=None ):
        return Response( {}, status=status.HTTP_404_NOT_FOUND )

    def retrieve( self, request, pk, format=None ):
        return Response( {}, status=status.HTTP_404_NOT_FOUND )

    def update( self, request, pk, format=None ):
        return Response( {}, status=status.HTTP_404_NOT_FOUND )


router = routers.SimpleRouter()
router.register( r'token', Mock_View_authenticated, basename='token' )
router.register( r'fail_soft', Mock_view_all_404, basename='fail_soft' )

urlpatterns = router.urls



from datetime import datetime
from django.urls import resolve
from rest_framework import status
import logging


logger = logging.getLogger( 'chibi_user.middleware' )


class User_for_api( object ):
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        self.process_request( request )
        response = self.get_response(request)
        return response

    def process_request( self, request ):
        request._start_time = datetime.utcnow()

    def process_template_response( self, request, response ):
        request._end_time = datetime.utcnow()
        request._total_time = request._start_time - request._end_time

        metrics = {
            'start_time': request._start_time,
            'end_time': request._end_time,
            'total_time': request._total_time,
        }
        self.set_user_to_metrics( request, metrics )
        self.is_an_error_from_user( request, response, metrics )

        logger.info( "metric for user", extra=metrics )

        return response

    def set_user_to_metrics( self, request, metrics ):
        if hasattr( request, 'user' ):
            metrics[ 'user_pk' ] = request.user.pk
        metrics[ 'is_anonymous' ] = 'user_pk' not in metrics

    def is_an_error_from_user( self, request, response, metrics ):
        url_name = self.get_url_name_from_response( request )
        if ( url_name and 'detail' in url_name ):
            metrics[ 'user_error' ] = \
                response.status_code == status.HTTP_404_NOT_FOUND
        else:
            metrics[ 'user_error' ] = False

    def get_url_name_from_response( self, request ):
        url_resolve = resolve( request.get_full_path() )
        return url_resolve.url_name



def batch( qs, batch_size=1000 ):
    """
    Returns a ( start, end, total, queryset ) tuple for each batch in the given
    queryset.

    Usage:
        # Make sure to order your querset
        article_qs = Article.objects.order_by('id')
        for start, end, total, qs in batch_qs(article_qs):
            print "Now processing %s - %s of %s" % (start + 1, end, total)
            for article in qs:
                print article.body
    """
    total = qs.count()
    for start in range( 0, total, batch_size ):
        end = min( start + batch_size, total )
        yield ( start, end, total, qs[ start:end ] )



import os
from unittest.suite import TestSuite

from django.conf import settings
from django.test.runner import DiscoverRunner


# sys.stdout = None
# requests.packages.urllib3.disable_warnings( InsecureRequestWarning )


class CustomizedRunner( DiscoverRunner ):
    def __init__( self, *args, **kw ):
        super().__init__( *args, **kw )

    def build_suite(self, *args, **kwargs):
        suite = super().build_suite( *args, **kwargs )
        filtered = TestSuite()

        for test in suite:
            testname = str( test )
            if '.tests.' in testname and self.package in testname:
                filtered.addTest( test )
        return filtered

    def setup_test_environment( self, **kargs ):
        settings.TEST_MODE = True
        super().setup_test_environment( **kargs )
        settings.CELERY_ALWAYS_EAGER = True
        settings.CELERY_EAGER_PROPAGATES_EXCEPTIONS = True  # Issue #75
        settings.DEBUG = False


class UnitRunner( CustomizedRunner ):
    package = '.unit.'

    def setup_databases( self, *args, **kwargs ):
        pass

    def teardown_databases( self, *args, **kwargs ):
        pass


class IntegrationRunner( CustomizedRunner ):
    package = '.integration.'


class AcceptanceRunner( CustomizedRunner ):
    package = '.acceptance.'
    os.environ[ 'DJANGO_LIVE_TEST_SERVER_ADDRESS' ] = 'localhost:8001'



from rest_framework.test import APIClient
from chibi_user.tests import get_user_test as helper_get_user_test
from chibi import madness
from unittest import TestCase
from chibi.snippet.string import decode
from pprint import pformat
from chibi.parser import link_header
from chibi.fancy.is_type import is_iter
import json


tc = TestCase( '__init__' )
default_client = APIClient()


def change_user( user_pk=None ):
    default_client = APIClient()
    user_test, token_test = helper_get_user_test( user_pk )
    default_client.credentials( HTTP_AUTHORIZATION=str( token_test ) )

try:
    change_user()
except Exception as e:
    print( e )


def request_get_abs( url, filters=None, client=None, ):
    if client is None:
        client = default_client
    return client.get( url, filters )


def get_location( response, filters=None, client=None ):
    location = response[ 'location' ]
    return request_get_abs( location, filters, client )


def get_url_of_response( response ):
    return response.request[ 'PATH_INFO' ]


def get_link_header( response ):
    return link_header( response[ 'link' ] )


def assert_has_pages( response ):
    tc.assertTrue(
        response.has_header( 'link' ), 'should have links for the pages' )


def assert_has_next_page( response ):
    links = get_link_header( response )
    tc.assertIn(
        'next', links,
        "\nEl header link no tiene siguiente pagina\n{}".format( links ) )


def assert_has_prev_page( response ):
    links = get_link_header( response )
    tc.assertIn(
        'prev', links,
        "\nEl header link no tiene anterior pagina\n{}".format( links ) )


def assert_status_code( response, status_ok, print_headers=False ):
    try:
        response_data = response.data
    except AttributeError:
        response_data = madness.string.decode( response.content )
        response_data = json.loads( response_data )
    if print_headers:
        msg = ( "\nthe status code should be '{}' but is '{}'\n"
                "url: {}\n"
                "the data of the response is:\n{}"
                ).format( status_ok, response.status_code,
                          get_url_of_response( response ),
                          pformat( response_data ) )
    else:
        msg = ( "\nthe status code should be '{}' but is '{}'\n"
                "url: {}\n"
                "the data of the response is:\n{}"
                ).format( status_ok, response.status_code,
                          get_url_of_response( response ),
                          pformat( response_data ) )
    if is_iter( status_ok ):
        tc.assertIn( response.status_code, status_ok, msg )
    else:
        tc.assertEqual( response.status_code, status_ok, msg )


def assert_data( response, data, print_headers=False ):
    try:
        response_data = response.data
    except AttributeError:
        response_data = decode( response.content )
        response_data = json.loads( response_data )
    if print_headers:
        msg = (
            "\nthe data of the aswer is not the expected\n"
            "status code: {}\ndata:\n{}\n"
            "expected data:\n{}").format(
                response.status_code, pformat( response_data ),
                pformat( data ) )
    else:
        msg = (
            "\nthe data of the aswer is not the expected\n"
            "status code: {}\ndata:\n{}\n"
            "expected data:\n{}").format(
                response.status_code, pformat( response_data ),
                pformat( data ) )
    tc.assertEqual( response_data, data, msg )


def assert_data_subset( response, data, print_headers=False ):
    try:
        response_data = response.data
    except AttributeError:
        response_data = decode( response.content )
        response_data = json.loads( response_data )
    if print_headers:
        msg = (
            "\nthe data of the aswer is not the expected\n"
            "status code: {}\ndata:\n{}\n"
            "expected data:\n{}").format(
                response.status_code, pformat( response_data ),
                pformat( data ) )
    else:
        msg = (
            "\nthe data of the aswer is not the expected\n"
            "status code: {}\ndata:\n{}\n"
            "expected data:\n{}").format(
                response.status_code, pformat( response_data ),
                pformat( data ) )
    tc.assertDictContainsSubset( data, response_data, msg )



"""
Django settings for chibi_django project.

Generated by 'django-admin startproject' using Django 2.0.7.

For more information on this file, see
https://docs.djangoproject.com/en/2.0/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/2.0/ref/settings/
"""

import os

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/2.0/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = '$_t$-0cqvn*u#9dpa)uu-p1520ddrsn0nq+a9-j+*n2es@*$tf'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django_extensions',
    'rest_framework',
    'corsheaders',
    'chibi_django',
    'chibi_user',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'chibi_django.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

WSGI_APPLICATION = 'chibi_django.wsgi.application'


# Database
# https://docs.djangoproject.com/en/2.0/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}


# Password validation
# https://docs.djangoproject.com/en/2.0/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME':
            'django.contrib.auth.password_validation.'
            'UserAttributeSimilarityValidator',
    },
    {
        'NAME':
            'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME':
            'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME':
            'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]


# Internationalization
# https://docs.djangoproject.com/en/2.0/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True

AUTH_USER_MODEL = 'chibi_user.User'


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/2.0/howto/static-files/

STATIC_URL = '/static/'

REST_FRAMEWORK = {
    # Use Django's standard `django.contrib.auth` permissions,
    'DEFAULT_AUTHENTICATION_CLASSES': [
        'chibi_user.authentication.Token_simple_authentication',
        # 'rest_framework.authentication.BasicAuthentication',
        # 'rest_framework.authentication.SessionAuthentication',
    ],
    'DEFAULT_PERMISSION_CLASSES': (
        'rest_framework.permissions.IsAuthenticated',
    ),
    'DEFAULT_RENDERER_CLASSES': (
        'rest_framework.renderers.JSONRenderer',
        # 'rest_framework_xml.renderers.XMLRenderer',
        # 'rest_framework_yaml.renderers.YAMLRenderer',
    ),
    'DEFAULT_PARSER_CLASSES': (
        'rest_framework.parsers.JSONParser',
        # 'rest_framework_xml.parsers.XMLParser',
        # 'rest_framework_yaml.parsers.YAMLParser',
    ),
    'DEFAULT_CONTENT_NEGOTIATION_CLASS':
        'rest_framework.negotiation.DefaultContentNegotiation',
    'EXCEPTION_HANDLER': 'chibi_django.exceptions.generic_exception_handler',
    'NON_FIELD_ERRORS_KEY': 'detail',
    'DEFAULT_VERSIONING_CLASS':
        'rest_framework.versioning.AcceptHeaderVersioning',
    'DEFAULT_VERSION': 'howard the duck',
    'ALLOWED_VERSIONS': ( 'howard the duck', ),
    'VERSION_PARAM': 'issue',
    'TEST_REQUEST_DEFAULT_FORMAT': 'json',
    'DEFAULT_FILTER_BACKENDS': (
        'django_filters.rest_framework.DjangoFilterBackend', ),
    'DEFAULT_PAGINATION_CLASS':
        'chibi_django.pagination.Link_header_pagination',
    'PAGE_SIZE': 5,
}


SOCIAL_AUTH_AUTH0_SCOPE = [ 'openid', 'profile' ]


SOCIAL_AUTH_TRAILING_SLASH = False  # Remove end slash from routes
SOCIAL_AUTH_AUTH0_DOMAIN = 'dem-test.auth0.com'
SOCIAL_AUTH_AUTH0_KEY = 'PM2yfdSM8rmzxbAhfKbvc6dXPACZHQDN'
SOCIAL_AUTH_AUTH0_SECRET = (
    'XNdgTZjolhAgHkriw1gZ7Kime6VbZ-p8I5O0EWJHZ-JSrPztc2WNJVV4jnOkSosL'
)


AUTHENTICATION_BACKENDS = {
    'chibi_user.auth0_backend.Auth0',
    'django.contrib.auth.backends.ModelBackend'
}


LOGIN_URL = "/login/auth0"
LOGIN_REDIRECT_URL = "/dashboard"
LOGOUT_REDIRECT_URL = "/"



from django.test import TestCase, override_settings
from django.utils import six
from rest_framework import status
from rest_framework.test import APIClient

from chibi_user.models import Token
from chibi_user.tests import get_user_test


class BaseTokenAuthTests( object ):
    """Token authentication"""
    model = None
    url_name = '/token/'
    header_prefix = 'Token '

    def setUp( self ):
        self.csrf_client = APIClient( enforce_csrf_checks=True )
        self.user, self.token = get_user_test()
        self.username = self.user.username
        self.key = self.token.key
        # self.path = reverse( self.url_name )

    def test_post_form_passing_token_auth( self ):
        """
        Ensure POSTing json over token auth with correct
        credentials passes and does not require CSRF
        """
        auth = self.header_prefix + self.key
        response = self.csrf_client.post(
            self.path, { 'example': 'example' }, HTTP_AUTHORIZATION=auth
        )
        self.assertEqual( response.status_code, status.HTTP_200_OK )

    def test_fail_post_form_passing_nonexistent_token_auth( self ):
        # use a nonexistent token key
        auth = self.header_prefix + 'wxyz6789'
        response = self.csrf_client.post(
            self.path, { 'example': 'example' }, HTTP_AUTHORIZATION=auth
        )
        self.assertEqual( response.status_code, status.HTTP_401_UNAUTHORIZED )

    def test_fail_post_form_passing_invalid_token_auth( self ):
        # add an 'invalid' unicode character
        auth = self.header_prefix + self.key + "¸"
        response = self.csrf_client.post(
            self.path, { 'example': 'example' }, HTTP_AUTHORIZATION=auth
        )
        self.assertEqual( response.status_code, status.HTTP_401_UNAUTHORIZED )

    def test_post_json_passing_token_auth( self ):
        """
        Ensure POSTing form over token auth with correct
        credentials passes and does not require CSRF
        """
        auth = self.header_prefix + self.key
        response = self.csrf_client.post(
            self.path, { 'example': 'example' },
            format='json', HTTP_AUTHORIZATION=auth
        )
        self.assertEqual( response.status_code, status.HTTP_200_OK )

    def test_post_json_makes_one_db_query( self ):
        """
        Ensure that authenticating a user using a
        token performs only one DB query
        """
        auth = self.header_prefix + self.key

        def func_to_test():
            return self.csrf_client.post(
                self.path, { 'example': 'example' },
                format='json', HTTP_AUTHORIZATION=auth
            )

        self.assertNumQueries( 1, func_to_test )

    def test_post_form_failing_token_auth( self ):
        """
        Ensure POSTing form over token auth without correct credentials fails
        """
        response = self.csrf_client.post( self.path, { 'example': 'example' } )
        self.assertEqual( response.status_code, status.HTTP_401_UNAUTHORIZED )

    def test_post_json_failing_token_auth( self ):
        """
        Ensure POSTing json over token auth without correct credentials fails
        """
        response = self.csrf_client.post(
            self.path, {'example': 'example'}, format='json'
        )
        self.assertEqual( response.status_code, status.HTTP_401_UNAUTHORIZED )


@override_settings( ROOT_URLCONF='chibi_user.tests' )
class TokenAuthTests( BaseTokenAuthTests, TestCase ):
    model = Token
    path = '/token/'

    def test_token_has_auto_assigned_key_if_none_provided( self ):
        """Ensure creating a token with no key will auto-assign a key"""
        self.token.delete()
        token = self.model.objects.create( user=self.user )
        self.assertTrue( bool( token.key ))

    def test_generate_key_returns_string( self ):
        """Ensure generate_key returns a string"""
        token = self.model()
        key = token.generate_key()
        self.assertIsInstance( key, six.string_types )



try:
    from setuptools import setup, find_packages
except:
    from distutils.core import setup, find_packages

# here = os.path.abspath( os.path.dirname( __file__ ) )
# README = open(os.path.join( here, 'README.rst' ) ).read()

setup(
    name='chibi_django',
    version='0.2.3',
    description='',
    # long_description=README,
    license='',
    author='',
    author_email='',
    packages=find_packages(include=[
        'chibi_django', 'chibi_django.*',
        'chibi_user', 'chibi_user.*',
        'test_runners', 'test_runners.*' ] ),
    install_requires=[
        'Django>=2.0.7', 'django-filter>=2.0.7',
        'djangorestframework>=3.8.2',
        'django-filters>=0.2.1', 'drf-nested-routers>=0.90.2',
        'chibi>=0.7.7', 'chibi_donkey==1.0.0'
    ],
    dependency_links = [],
    url='https://github.com/dem4ply/chibi_django',
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
    ],
)



from django.conf import settings


def build_index_name( name, app_name=None, ):
    if app_name is None:
        app_name = getattr( settings, 'PROJECT_NAME', None )
    if not app_name:
        result = name
    else:
        result = f"{app_name}__{name}"

    is_test = getattr( settings, 'TEST_MODE', False )
    if is_test:
        return f"test__{result}"
    return result



from rest_framework_nested import routers
from . import views

router = routers.SimpleRouter()

router.register( r'users', views.User, basename='users' )

token_router = routers.NestedSimpleRouter( router, r'users', lookup='users' )
token_router.register( r'token', views.Token, basename='tokens' )


urlpatterns = router.urls + token_router.urls



def show_urls( urllist, depth=0 ):
    for entry in urllist:
        if ( hasattr( entry, 'namespace' ) ):
            print( "\t" * depth, entry.pattern.regex.pattern,
                   "[%s]" % entry.namespace )
        else:
            print( "\t" * depth, entry.pattern.regex.pattern,
                   "[%s]" % entry.name )
        if hasattr( entry, 'url_patterns' ):
            show_urls( entry.url_patterns, depth + 1 )



from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APITestCase
from test_runners.snippet.response import assert_status_code


class Test_status_code( APITestCase ):
    expected_status_code = status.HTTP_200_OK
    url_name = 'namespace:name'


class Test_list( Test_status_code ):
    def test_list( self ):
        url_kw = self.get_url_kw()
        url = reverse( self.url_name, kwargs=url_kw )
        response = self.client.get( url )
        assert_status_code( response, self.expected_status_code )

    def get_url_kw( self ):
        raise NotImplementedError


class Test_retrieve( Test_status_code ):
    def test_retrieve( self ):
        url_kw = self.get_url_kw()
        url = reverse( self.url_name, kwargs=url_kw )
        response = self.client.get( url )
        assert_status_code( response, self.expected_status_code )

    def get_url_kw( self ):
        raise NotImplementedError

    def get_post_data( self ):
        raise NotImplementedError


class Test_create( Test_status_code ):
    expected_status_code = status.HTTP_201_CREATED

    def test_create( self ):
        url_kw = self.get_url_kw()
        url = reverse( self.url_name, kwargs=url_kw )
        post_data = self.get_post_data()
        response = self.client.post( url, data=post_data )
        assert_status_code( response, self.expected_status_code )

    def get_url_kw( self ):
        raise NotImplementedError

    def get_post_data( self ):
        raise NotImplementedError



# Generated by Django 2.0.7 on 2019-02-26 11:01

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0009_alter_user_last_name_max_length'),
    ]

    operations = [
        migrations.CreateModel(
            name='User',
            fields=[
                ('password', models.CharField(max_length=128, verbose_name='password')),
                ('last_login', models.DateTimeField(blank=True, null=True, verbose_name='last login')),
                ('id', models.CharField(max_length=64, primary_key=True, serialize=False, verbose_name='Id of user')),
                ('username', models.CharField(max_length=64, unique=True)),
                ('email', models.CharField(default=None, max_length=64, null=True, unique=True)),
                ('first_name', models.CharField(max_length=64)),
                ('last_name', models.CharField(max_length=64)),
                ('is_active', models.BooleanField(default=True)),
                ('is_staff', models.BooleanField(default=False)),
                ('is_superuser', models.BooleanField(default=False)),
                ('groups', models.ManyToManyField(blank=True, help_text='The groups this user belongs to. A user will get all permissions granted to each of their groups.', related_name='user_set', related_query_name='user', to='auth.Group', verbose_name='groups')),
                ('user_permissions', models.ManyToManyField(blank=True, help_text='Specific permissions for this user.', related_name='user_set', related_query_name='user', to='auth.Permission', verbose_name='user permissions')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Token',
            fields=[
                ('key', models.CharField(max_length=64, primary_key=True, serialize=False)),
                ('create_at', models.DateTimeField(auto_now_add=True)),
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='token', to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



from django.contrib import admin
from django.urls import path, include
from chibi_user import views as chibi_user_views
from chibi_user import urls as users_urls

urlpatterns = [
    path( '^$', chibi_user_views.index ),
    path( r'^dashboard', chibi_user_views.dashboard ),
    path(
        r'^', include(
            ( 'django.contrib.auth.urls', 'django_contrib' ),
            namespace='auth' ) ),
    path( 'admin/', admin.site.urls ),
    path( r'', include( ( users_urls, 'users' ), namespace='users' ), ),
]



from .models import User as User_model, Token as Token_model
from .serializers import (
    User as User_serializer, Token as Token_serializer,
    User_create as User_create_serializer
)
from .authentication import Token_simple_authentication
from rest_framework.decorators import action
from rest_framework.permissions import IsAdminUser
from rest_framework.response import Response
from chibi_django import view_set
from rest_framework import filters
from django.shortcuts import render


from django.contrib.auth.decorators import login_required

import json


def index( request ):
    return render( request, 'index.html' )


@login_required
def dashboard( request ):
    user = request.user
    auth0user = user.social_auth.get( provider="auth0")
    userdata = {
        'user_id': auth0user.uid,
        'name': user.first_name,
        'picture': auth0user.extra_data['picture']
    }

    return render(
        request, 'dashboard.html',
        {
            'auth0User': auth0user,
            'userdata': json.dumps( userdata, indent=4 )
        } )


class User( view_set.Model_viewset ):
    authentication_classes = [ Token_simple_authentication ]
    permission_classes = [ IsAdminUser ]
    queryset = User_model.objects.all()
    filter_backends = ( filters.OrderingFilter, )
    ordering_fields = ( 'username', 'email' )
    ordering = ( 'username', )
    serializer_class = {
        'default': User_serializer,
        'create': User_create_serializer,
    }

    @action( detail=True, methods=[ 'POST' ] )
    def refresh_token( self, request, pk, format=None ):
        user = User_model.objects.get( pk=pk )
        token = user.refresh_token()
        serializer = Token_serializer( token )
        return Response( serializer.data )


class Token( view_set.Model_viewset ):
    authentication_classes = [ Token_simple_authentication ]
    permission_classes     = [ IsAdminUser ]
    # queryset               = Token_model.objects.all()
    serializer_class       = Token_serializer

    def get_queryset( self, *args, **kw ):
        user_pk = self.kwargs[ 'users_pk' ]
        return Token_model.objects.filter( user__pk=user_pk )



from django.db import connection
from django.db.models.base import ModelBase
from chibi_django.models import Chibi_model
from django.test import TestCase


class ModelMixinTestCase( TestCase ):
    @classmethod
    def setUpClass(cls):
        cls.model = ModelBase(
            '__TestModel__' + cls.model.__name__, ( cls.model, ),
            { '__module__': cls.model.__module__ }
        )

        with connection.schema_editor() as schema_editor:
            schema_editor.create_model(cls.model)
        super(ModelMixinTestCase, cls).setUpClass()


class Test_model( Chibi_model ):
    pass


class Test_chibi_model( ModelMixinTestCase ):
    model = Test_model

    def test_model_using_save( self ):
        model = Test_model()
        self.assertFalse( model.pk )
        model.save()
        model.refresh_from_db()
        self.assertTrue( model.pk )

    def test_model_using_save_twice_should_no_change_the_pk( self ):
        model = Test_model()
        self.assertFalse( model.pk )
        model.save()
        model.refresh_from_db()
        old_pk = model.pk
        model.save()
        self.assertEqual( old_pk, model.pk )

    def test_model_with_create( self ):
        model = Test_model.objects.create()
        self.assertTrue( model.pk )
        model.refresh_from_db()
        old_pk = model.pk
        model.save()
        self.assertEqual( old_pk, model.pk )
